import logging

# Dictionary of logging levels of corresponding classes
logging_levels = {
                'ChromatecControlPanelModbus'               	:		logging.INFO,
                'ChromatecAnalyticModbus'						:		logging.INFO,
                'ChromatecCrystal5000'							:		logging.INFO,
                'pycatalicism.chromatograph.modbus_converter'	:		logging.INFO,
                }
