class MFCConnectionException(Exception):
    """
    Class represents exceptions related to connection errors
    """

class MFCStateException(Exception):
    """
    Class represents exceptions related to wrong state of the device
    """
