import logging

# Dictionary of logging levels of corresponding classes
logging_levels = {
                'OwenProtocol'	:	logging.INFO,
                'OwenTPM101'	:	logging.INFO,
                }
